export declare type AnyFunction = (...args: any[]) => any;

export declare type App<T extends AppBase = AppBase> = {
    schema: Project;
    config: PlainObject;
    readonly boosts: AppBoosts;
    use(plugin: Plugin_2): Promise<void>;
} & T;

export declare interface AppBase {
    mount: (el: HTMLElement) => void | Promise<void>;
    unmount: () => void | Promise<void>;
}

declare interface AppBoosts {
    [key: string]: any;
}

declare interface AppBoostsManager {
    hookStore: HookStore<RuntimeHooks>;
    readonly value: AppBoosts;
    add(name: PropertyKey, value: any, force?: boolean): void;
    remove(name: PropertyKey): void;
}

/**
 * context for plugin or renderer
 */
export declare interface AppContext {
    schema: AppSchema;
    config: PlainObject;
    appScope: CodeScope;
    packageManager: PackageManager;
    boosts: AppBoostsManager;
}

declare type AppCreator<O, T extends AppBase> = (appContext: Omit<AppContext, 'renderer'>, appOptions: O) => Promise<{
    appBase: T;
    renderer?: any;
}>;

export declare interface AppOptionsBase {
    schema: Project;
    packages?: Package[];
    plugins?: Plugin_2[];
    appScopeValue?: PlainObject;
}

declare interface AppSchema {
    getComponentsTrees(): ComponentTree[];
    addComponentsTree(tree: ComponentTree): void;
    removeComponentsTree(id: string): void;
    getComponentsMaps(): ComponentMap[];
    addComponentsMap(componentName: ComponentMap): void;
    removeComponentsMap(componentName: string): void;
    getPageConfigs(): PageConfig[];
    addPageConfig(page: PageConfig): void;
    removePageConfig(id: string): void;
    getByKey<K extends keyof Project>(key: K): Project[K] | undefined;
    updateByKey<K extends keyof Project>(key: K, updater: Project[K] | ((value: Project[K]) => Project[K])): void;
    getByPath(path: string | string[]): any;
    updateByPath(path: string | string[], updater: any | ((value: any) => any)): void;
    find(predicate: (schema: Project) => any): any;
}

export declare interface CodeRuntime {
    run<T = unknown>(code: string): T | undefined;
    createFnBoundScope(code: string): AnyFunction | undefined;
    parseExprOrFn(value: PlainObject): any;
    bindingScope(scope: CodeScope): void;
    getScope(): CodeScope;
}

export declare interface CodeScope<T extends PlainObject = PlainObject, K extends keyof T = keyof T> {
    readonly value: T;
    inject(name: K, value: T[K], force?: boolean): void;
    setValue(value: T, replace?: boolean): void;
    createSubScope<O extends PlainObject = PlainObject>(initValue: O): CodeScope<T & O>;
}

/**
 * 复杂 urls 结构，同时兼容简单结构和多模态结构
 */
export declare type ComplexUrls = string[] | MultiModeUrls;

/**
 * https://lowcode-engine.cn/site/docs/specs/lowcode-spec#22-%E7%BB%84%E4%BB%B6%E6%98%A0%E5%B0%84%E5%85%B3%E7%B3%BBa
 * 协议中用于描述 componentName 到公域组件映射关系的规范。
 */
export declare interface ComponentMap {
    /**
     * 协议中的组件名，唯一性，对应包导出的组件名，是一个有效的 JS 标识符，而且是大写字母打头
     */
    componentName: string;
    /**
     * npm 公域的 package name
     */
    package?: string;
    /**
     * package version
     */
    version?: string;
    /**
     * 使用解构方式对模块进行导出
     */
    destructuring?: boolean;
    /**
     * 包导出的组件名
     */
    exportName?: string;
    /**
     * 下标子组件名称
     */
    subName?: string;
    /**
     * 包导出组件入口文件路径
     */
    main?: string;
    /**
     * proCode or lowCode
     */
    devMode?: string;
}

/**
 * 组件树描述
 * 协议中用于描述搭建出来的组件树结构的规范，整个组件树的描述由组件结构&容器结构两种结构嵌套构成。
 */
export declare type ComponentTree<LifeCycleNameT extends string = string> = ComponentTreeContainer<LifeCycleNameT>;

/**
 * 容器结构描述 (A)
 * 容器是一类特殊的组件，在组件能力基础上增加了对生命周期对象、自定义方法、样式文件、数据源等信息的描述。
 */
export declare interface ComponentTreeContainer<LifeCycleNameT extends string> extends Omit<ComponentTreeNode, 'loop' | 'loopArgs' | 'condition'> {
    componentName: 'Page' | 'Block' | 'Component';
    /**
     * 文件名称
     */
    fileName: string;
    /**
     * 容器初始数据
     */
    state?: Record<string, JSONValue | JSExpression>;
    /**
     * 样式属性
     */
    css?: string;
    /**
     * 生命周期对象
     */
    lifeCycles?: {
        [name in LifeCycleNameT]: JSFunction;
    };
    /**
     * 自定义方法对象
     */
    methods?: {
        [name: string]: JSFunction;
    };
    /**
     * 数据源对象
     * type todo
     */
    dataSource?: any;
}

/**
 * 组件结构描述（A）
 */
export declare interface ComponentTreeNode {
    /**
     * 组件唯一标识
     */
    id?: string;
    /**
     * 组件名称
     */
    componentName: string;
    /**
     * 组件属性对象
     */
    props?: ComponentTreeNodeProps;
    /**
     * 选填，根据表达式结果判断是否渲染物料；
     */
    condition?: boolean | JSExpression;
    /**
     * 循环数据
     */
    loop?: unknown[] | JSExpression;
    /**
     * 循环迭代对象、索引名称，默认为 ["item", "index"]
     */
    loopArgs?: [string, string];
    /**
     * 子组件
     */
    children?: NodeType[];
}

/**
 * Props 结构描述
 */
export declare interface ComponentTreeNodeProps {
    /** 组件 ID	 */
    id?: string | JSExpression;
    /** 组件样式类名	 */
    className?: string;
    /**  组件内联样式	 */
    style?: JSONObject | JSExpression;
    /** 组件 ref 名称	 */
    ref?: string;
    [key: string]: any;
}

export declare class ComponentWidget<E = unknown> extends Widget<ComponentTreeNode, E> {
    private _children;
    private _propsValue;
    protected init(): void;
    get componentName(): string;
    get props(): ComponentTreeNodeProps;
    get condition(): boolean;
    get loop(): unknown[] | JSExpression | undefined;
    get loopArgs(): [string, string];
    get children(): (ComponentWidget<E> | TextWidget<E>)[];
}

/**
 * 应用范围内的全局常量（AA）
 * 用于描述在整个应用内通用的全局常量，比如请求 API 的域名、环境等。
 */
export declare interface ConstantsMap {
    [key: string]: JSONValue;
}

/**
 * 根据低代码搭建协议的容器组件描述生成的容器实例
 */
export declare interface Container<InstanceT = unknown, LifeCycleNameT extends string = string> {
    readonly codeRuntime: CodeRuntime;
    readonly instanceApiObject: InstanceApi<InstanceT>;
    /**
     * 获取协议中的 css 内容
     */
    getCssText(): string | undefined;
    /**
     * 调用生命周期方法
     */
    triggerLifeCycle(lifeCycleName: LifeCycleNameT, ...args: any[]): void;
    /**
     * 设置 ref 对应的组件实例, 提供给 scope.$() 方式使用
     */
    setInstance(ref: string, instance: InstanceT): void;
    /**
     * 移除 ref 对应的组件实例
     */
    removeInstance(ref: string, instance?: InstanceT): void;
    createWidgets<Element>(): (TextWidget<Element> | ComponentWidget<Element>)[];
}

/**
 * 创建 createApp 的辅助函数
 * @param schema
 * @param options
 * @returns
 */
export declare function createAppFunction<O extends AppOptionsBase, T extends AppBase = AppBase>(appCreator: AppCreator<O, T>): (options: O) => Promise<App<T>>;

export declare function createCodeRuntime(scopeOrValue?: PlainObject): CodeRuntime;

export declare type CreateComponentBaseOptions<T extends string> = Omit<CreateContainerOptions<T>, 'stateCreator'>;

/**
 * 创建 createComponent 的辅助函数
 * createComponent = createComponentFunction(() => component)
 */
export declare function createComponentFunction<ComponentT, InstanceT, LifeCycleNameT extends string, O extends CreateComponentBaseOptions<LifeCycleNameT>>(stateCreator: (initState: PlainObject) => InstanceStateApi, componentCreator: (container: Container<InstanceT, LifeCycleNameT>, componentOptions: O) => ComponentT): (componentOptions: O) => ComponentT;

export declare function createContainer<InstanceT, LifeCycleNameT extends string>(options: CreateContainerOptions<LifeCycleNameT>): Container<InstanceT, LifeCycleNameT>;

export declare interface CreateContainerOptions<LifeCycleNameT extends string> {
    supCodeScope?: CodeScope;
    initScopeValue?: PlainObject;
    componentsTree: ComponentTree<LifeCycleNameT>;
    stateCreator: (initalState: PlainObject) => InstanceStateApi;
    dataSourceCreator: (...args: any[]) => InstanceDataSourceApi;
}

export declare function createEvent<T = AnyFunction>(): {
    add: (fn: T) => () => void;
    remove: (fn: T) => void;
    list: () => T[];
    clear(): void;
};

export declare function createHookStore<HooksT extends Record<PropertyKey, any> = Record<PropertyKey, HookCallback>, HookNameT extends HookKeys<HooksT> = HookKeys<HooksT>>(): HookStore<HooksT, HookNameT>;

export declare function createScope<T extends PlainObject = PlainObject, K extends keyof T = keyof T>(initValue: T): CodeScope<T, K>;

export declare function createWidget<E = unknown>(data: NodeType): TextWidget<E> | ComponentWidget<E>;

export declare function definePlugin<C extends PluginSetupContext, P = Plugin_2<C>>(plugin: P): P;

declare type ErrorType = string;

export declare interface FunctionUtil {
    name: string;
    type: 'function';
    content: JSFunction;
}

declare type HookCallback = (...args: any) => Promise<any> | any;

declare type HookKeys<T> = keyof T & PropertyKey;

declare interface HookStore<HooksT extends Record<PropertyKey, any> = Record<PropertyKey, HookCallback>, HookNameT extends HookKeys<HooksT> = HookKeys<HooksT>> {
    hook<NameT extends HookNameT>(name: NameT, fn: InferCallback<HooksT, NameT>): () => void;
    call<NameT extends HookNameT>(name: NameT, ...args: Parameters<InferCallback<HooksT, NameT>>): void;
    callAsync<NameT extends HookNameT>(name: NameT, ...args: Parameters<InferCallback<HooksT, NameT>>): Promise<void>;
    callParallel<NameT extends HookNameT>(name: NameT, ...args: Parameters<InferCallback<HooksT, NameT>>): Promise<void[]>;
    remove<NameT extends HookNameT>(name: NameT, fn?: InferCallback<HooksT, NameT>): void;
    clear<NameT extends HookNameT>(name?: NameT): void;
    getHooks<NameT extends HookNameT>(name: NameT): InferCallback<HooksT, NameT>[] | undefined;
}

/**
 * https://lowcode-engine.cn/site/docs/specs/lowcode-spec#25-%E5%9B%BD%E9%99%85%E5%8C%96%E5%A4%9A%E8%AF%AD%E8%A8%80%E6%94%AF%E6%8C%81aa
 * 国际化多语言支持
 */
export declare interface I18nMap {
    [locale: string]: Record<string, string>;
}

/**
 * 国际化多语言类型（AA）
 */
export declare interface I18nNode {
    type: 'i18n';
    /**
     * i18n 结构中字段的 key 标识符
     */
    key: string;
    /**
     * 语料为字符串模板时的变量内容
     */
    params?: Record<string, string | number | JSExpression>;
}

declare type InferCallback<HT, HN extends keyof HT> = HT[HN] extends HookCallback ? HT[HN] : never;

/**
 * 在上述事件类型描述和变量类型描述中，在函数或 JS 表达式内，均可以通过 this 对象获取当前组件所在容器的实例化对象
 * 在搭建场景下的渲染模块和出码模块实现上，统一约定了该实例化 this 对象下所挂载的最小 API 集合，
 * 以保障搭建协议具备有一致的数据流和事件上下文
 */
export declare interface InstanceApi<InstanceT = unknown> extends InstanceStateApi, InstanceDataSourceApi {
    /**
     * 容器的 props 对象
     */
    props?: PlainObject;
    /**
     * ref 对应组件上配置的 ref 属性，用于唯一标识一个组件；若有同名的，则会返回第一个匹配的。
     * @param ref 组件标识
     */
    $(ref: string): InstanceT | undefined;
    /**
     * ref 对应组件上配置的 ref 属性，用于唯一标识一个组件；总是返回一个数组，里面是所有匹配 ref 的组件的引用。
     * @param ref 组件标识
     */
    $$(ref: string): InstanceT[];
    [methodName: string]: any;
}

export declare interface InstanceDataSourceApi {
    /**
     * 实例的数据源对象 Map
     */
    dataSourceMap: any;
    /**
     * 实例的初始化异步数据请求重载
     */
    reloadDataSource: () => void;
}

export declare interface InstanceStateApi<S = PlainObject> {
    /**
     * 实例的数据对象 state
     */
    state: Readonly<S>;
    /**
     * 实例更新数据的方法
     * like React.Component.setState
     */
    setState<K extends keyof S>(newState: ((prevState: Readonly<S>) => Pick<S, K> | S | null) | (Pick<S, K> | S | null), callback?: () => void): void;
}

/**
 * 国际化相关 API
 */
export declare interface IntlApi {
    /**
     * 返回语料字符串
     * @param i18nKey 语料的标识符
     * @param params 可选，是用来做模版字符串替换
     */
    i18n(i18nKey: string, params?: Record<string, string>): string;
    /**
     * 返回当前环境语言
     */
    getLocale(): string;
    /**
     * 设置当前环境语言
     * @param locale 环境语言
     */
    setLocale(locale: string): void;
}

export declare function isI18nNode(v: unknown): v is I18nNode;

export declare function isJSExpression(v: unknown): v is JSExpression;

export declare function isJSFunction(v: unknown): v is JSFunction;

export declare function isJSSlot(v: unknown): v is JSSlot;

export declare function isLowCodeComponentSchema(v: unknown): v is LowCodeComponent;

/**
 * 变量类型（A）
 */
export declare interface JSExpression {
    type: 'JSExpression';
    value: string;
}

/**
 * 事件函数类型（A）
 */
export declare interface JSFunction {
    type: 'JSFunction';
    value: string;
}

export declare interface JSONObject {
    [key: string]: JSONValue | JSONObject | JSONObject[];
}

export declare type JSONValue = number | string | boolean | null;

/**
 * 节点类型（A）
 * 通常用于描述组件的某一个属性为 ReactNode 或 Function-Return-ReactNode 的场景。
 */
export declare interface JSSlot {
    type: 'JSSlot';
    value: ComponentTreeNode | ComponentTreeNode[];
    params?: string[];
}

/**
 * 资源加载环境种类
 */
export declare type LoadEnv = 'design' | 'runtime';

export declare interface LowCodeComponent extends Omit<Package, 'schema'> {
    id: string;
    type: 'lowCode';
    componentName: string;
    schema: ComponentTree;
}

/**
 * 多模态资源
 */
export declare interface MultiModeUrls {
    /**
     * 默认的资源 url
     */
    default: string[];
    /**
     * 其他模态资源的 url
     */
    [mode: string]: string[];
}

export declare type NodeType = string | JSExpression | I18nNode | ComponentTreeNode;

export declare interface NPMUtil {
    name: string;
    type: 'npm';
    content: Omit<ComponentMap, 'componentName'>;
}

export declare interface Package {
    /**
     * 唯一标识，与 package 必须要有一个存在值，如果为空，则以 package 为唯一标识
     */
    id?: string;
    /**
     * npm 包唯一标识，与 id 必须要有一个存在值
     */
    package?: string;
    /**
     * 包版本号
     */
    version: string;
    /**
     * 资源类型，默认为 proCode
     */
    type?: 'proCode' | 'lowCode';
    /**
     * 组件渲染态视图打包后的 CDN url 列表，包含 js 和 css
     */
    urls?: string[];
    /**
     * 组件多个渲染态视图打包后的 CDN url 列表，包含 js 和 css，优先级高于 urls
     */
    advancedUrls?: ComplexUrls;
    /**
     * 组件编辑态视图打包后的 CDN url 列表，包含 js 和 css
     */
    editUrls?: string[];
    /**
     * 组件多个编辑态视图打包后的 CDN url 列表，包含 js 和 css，优先级高于 editUrls
     */
    advancedEditUrls?: ComplexUrls;
    /**
     * 低代码组件的 schema 内容
     */
    schema?: Project;
    /**
     * 当前资源所依赖的其他资源包的 id 列表
     */
    deps?: string[];
    /**
     * 指定当前资源加载的环境
     */
    loadEnv?: LoadEnv[];
    /**
     * 当前资源是否是 external 资源
     */
    external?: boolean;
    /**
     * 作为全局变量引用时的名称，和 webpack output.library 字段含义一样，用来定义全局变量名
     */
    library: string;
    /**
     * 组件描述导出名字，可以通过 window[exportName] 获取到组件描述的 Object 内容；
     */
    exportName?: string;
    /**
     * 标识当前 package 资源加载在 window.library 上的是否是一个异步对象
     */
    async?: boolean;
    /**
     * 标识当前 package 从其他 package 的导出方式
     */
    exportMode?: string;
    /**
     * 标识当前 package 内容是从哪个 package 导出来的
     */
    exportSourceId?: string;
    /**
     * 标识当前 package 是从 window 上的哪个属性导出来的
     */
    exportSourceLibrary?: string;
}

export declare interface PackageLoader {
    name?: string;
    load(packageInfo: Package, thisManager: PackageManager): Promise<any>;
    active(packageInfo: Package): boolean;
}

export declare interface PackageManager {
    /**
     * 新增资产包
     * @param packages
     */
    addPackages(packages: Package[]): Promise<void>;
    /** 通过包名获取资产包信息 */
    getPackageInfo(packageName: string): Package | undefined;
    getLibraryByPackageName(packageName: string): any;
    setLibraryByPackageName(packageName: string, library: any): void;
    /** 新增资产包加载器 */
    addPackageLoader(loader: PackageLoader): void;
    /** 解析组件映射 */
    resolveComponentMaps(componentMaps: ComponentMap[]): void;
    /** 获取组件映射对象，key = componentName value = component */
    getComponentsNameRecord<C = unknown>(componentMaps?: ComponentMap[]): Record<string, C | LowCodeComponent>;
    /** 通过组件名获取对应的组件  */
    getComponent<C = unknown>(componentName: string): C | LowCodeComponent | undefined;
    /** 注册组件 */
    registerComponentByName(componentName: string, Component: unknown): void;
}

/**
 * 当前应用的页面信息（AA）
 * 用于描述当前应用的页面信息，比如页面对应的低代码搭建内容、页面标题、页面配置等。
 * 在一些比较复杂的场景下，声明一层页面映射关系，以支持页面声明更多信息与配置，同时能够支持不同类型的产物。
 */
export declare interface PageConfig {
    /**
     * 页面 id
     */
    id: string;
    /**
     * 页面类型，如 componentsTree，package，默认为 componentsTree
     */
    type?: string;
    /**
     * 映射的 id，根据页面的类型 type 判断及确定目标的 id
     */
    mappingId: string;
    /**
     * 页面元信息
     */
    meta?: JSONObject;
    /**
     * 页面配置
     */
    config?: JSONObject;
}

export declare type PlainObject = Record<PropertyKey, any>;

declare interface Plugin_2<C extends PluginSetupContext = PluginSetupContext> {
    name: string;
    setup(setupContext: C): void | Promise<void>;
    dependsOn?: string[];
}
export { Plugin_2 as Plugin }

export declare interface PluginSetupContext extends AppContext {
    [key: string]: any;
}

export declare function processValue(obj: any, predicate: (obj: any) => boolean, processor: (node: any, paths: Array<string | number>) => any): any;

export declare interface ProCodeComponent extends Package {
    package: string;
    type: 'proCode';
}

/**
 * https://lowcode-engine.cn/site/docs/specs/lowcode-spec
 * 低代码引擎搭建协议规范
 */
/**
 * https://lowcode-engine.cn/site/docs/specs/lowcode-spec#2-%E5%8D%8F%E8%AE%AE%E7%BB%93%E6%9E%84
 * 应用协议
 */
export declare interface Project {
    /**
     * 当前协议版本号
     */
    version: string;
    /**
     * 组件映射关系
     */
    componentsMap: ComponentMap[];
    /**
     * 描述模版/页面/区块/低代码业务组件的组件树
     */
    componentsTree: ComponentTree[];
    /**
     * 工具类扩展映射关系
     */
    utils?: Util[];
    /**
     * 国际化语料
     */
    i18n?: I18nMap;
    /**
     * 应用范围内的全局常量
     */
    constants?: ConstantsMap;
    /**
     * 应用范围内的全局样式
     * 用于描述在应用范围内的全局样式，比如 reset.css 等。
     */
    css?: string;
    /**
     * 当前应用配置信息
     */
    config?: Record<string, JSONValue>;
    /**
     * 当前应用元数据信息
     */
    meta?: Record<string, JSONValue>;
    /**
     * 当前应用的公共数据源
     * @deprecated
     */
    dataSource?: unknown;
    /**
     * 当前应用的路由配置信息
     */
    router?: RouterConfig;
    /**
     * 当前应用的所有页面信息
     */
    pages?: PageConfig[];
}

export declare type RawLocation = RawLocationAsPath | RawLocationAsRelative | RawLocationAsName;

export declare interface RawLocationAsName {
    name: string;
    params?: Record<string, string | string[]>;
}

export declare interface RawLocationAsPath {
    path: string;
}

export declare interface RawLocationAsRelative {
    params?: Record<string, string | string[]>;
}

export declare interface RawLocationOptions {
    searchParams?: URLSearchParams;
    hash?: string;
    state?: History['state'];
}

/**
 * 允许用户输入的路径参数类型
 */
export declare type RawRouteLocation = string | (RawLocation & RawLocationOptions);

/**
 * 路由的当前信息，模拟 window.location
 */
export declare interface RouteLocation {
    /**
     * 匹配到的路由记录名
     */
    name: string | undefined;
    /**
     * 当前解析后的路径
     */
    path: string;
    /**
     * 当前路径的 hash 值，以 # 开头
     */
    hash: string;
    /**
     * 匹配到的路径参数
     */
    params: Record<string, string | string[]> | undefined;
    /**
     * 当前的路径 URLSearchParams 对象
     */
    searchParams: URLSearchParams | undefined;
    /**
     * 包括 search 和 hash 在内的完整地址
     */
    fullPath: string;
    /**
     * 匹配到的路由记录元数据
     */
    meta: PlainObject | undefined;
    /**
     * 重定向之前的路由，在跳转到当前路径之前的路由记录
     */
    redirectedFrom: RouteLocation | undefined;
}

/**
 * 路由 Router API：封装了原生的 History、Location 等 api，提供统一的调用方法
 * 得益于 HTML 5 新的 History api 的规范出现，SPA 大行其道，其中 SPA 的路由起到了非常重大的作用
 */
export declare interface RouterApi {
    /**
     * 获取当前解析后的路由信息
     */
    getCurrentLocation(): RouteLocation;
    /**
     * 路由跳转方法，跳转到指定的路径或者 `Route`
     */
    push(location: RawRouteLocation): void | Promise<void>;
    /**
     * 路由跳转方法，与 `push` 的区别在于不会增加一条历史记录而是替换当前的历史记录
     */
    replace(location: RawRouteLocation): void | Promise<void>;
    /**
     * 返回上一页，同 `history.back`
     */
    back(): void;
    /**
     * 跳转下一页，同 `history.forward`
     */
    forward(): void;
    /**
     * 跳转到当前页面的相对位置，同 `history.go`
     * @param delta 相对于当前页面你要去往历史页面的位置
     */
    go(delta: number): void;
    /**
     * 路由跳转前的守卫方法
     */
    beforeRouteLeave(guard: (to: RouteLocation, from: RouteLocation) => boolean | Promise<boolean>): () => void;
    /**
     * 路由跳转后的钩子函数
     */
    afterRouteChange(guard: (to: RouteLocation, from: RouteLocation) => any): () => void;
}

/**
 * https://lowcode-engine.cn/site/docs/specs/lowcode-spec#211-%E5%BD%93%E5%89%8D%E5%BA%94%E7%94%A8%E7%9A%84%E8%B7%AF%E7%94%B1%E4%BF%A1%E6%81%AFaa
 * 当前应用的路由信息（AA）
 * 用于描述当前应用的路径 - 页面的关系。通过声明路由信息，应用能够在不同的路径里显示对应的页面。
 */
export declare interface RouterConfig {
    /**
     * 应用根路径
     */
    baseName: string;
    /**
     * 路由的 history 模式
     */
    historyMode: 'browser' | 'hash' | 'memory';
    /**
     * 路由对象组，路径与页面的关系对照组
     */
    routes: RouteRecord[];
}

/**
 * Route （路由记录）结构描述
 * 路由记录，路径与页面的关系对照。Route 的结构说明：
 */
export declare interface RouteRecord {
    /**
     * 该路径项的名称
     */
    name?: string;
    /**
     * 路径
     */
    path: string;
    /**
     * 路径对应的页面 ID，page 与 redirect 字段中必须要有有一个存在
     */
    page?: string;
    /**
     * 此路径需要重定向到的路由信息，page 与 redirect 字段中必须要有有一个存在
     */
    redirect?: string | object | JSFunction;
    /**
     * 子路由
     */
    children?: RouteRecord[];
}

declare class RuntimeError extends Error {
    type: ErrorType;
    constructor(type: ErrorType, message: string);
}

declare interface RuntimeHooks {
    'app:error': (error: RuntimeError) => void;
    [key: PropertyKey]: AnyFunction;
}

export declare function someValue(obj: any, predicate: (data: any) => boolean): boolean;

export declare class TextWidget<E = unknown> extends Widget<TextWidgetData, E> {
    type: TextWidgetType;
    protected init(): void;
}

export declare type TextWidgetData = Exclude<NodeType, ComponentTreeNode>;

export declare type TextWidgetType = 'string' | 'expression' | 'i18n';

/**
 * https://lowcode-engine.cn/site/docs/specs/lowcode-spec#24-%E5%B7%A5%E5%85%B7%E7%B1%BB%E6%89%A9%E5%B1%95%E6%8F%8F%E8%BF%B0aa
 * 用于描述物料开发过程中，自定义扩展或引入的第三方工具类（例如：lodash 及 moment），增强搭建基础协议的扩展性，提供通用的工具类方法的配置方案及调用 API。
 */
export declare type Util = NPMUtil | FunctionUtil;

/**
 * 应用级别的公共函数或第三方扩展
 */
export declare interface UtilsApi {
    utils: Record<string, AnyFunction>;
}

export declare class Widget<Data, Element> {
    raw: Data;
    protected proxyElements: Element[];
    protected renderObject: Element | undefined;
    constructor(raw: Data);
    protected init(): void;
    get key(): string;
    mapRenderObject(mapper: (widget: Widget<Data, Element>) => Element | undefined): this;
    addProxyELements(el: Element): void;
    build<C>(builder: (elements: Element[]) => C): C;
}

export { }
